// This is NOT an official Andor tool. The rights refering the game system and trademark Andor belong to Michael Menzel.
//
// The following described MIT-license applies ONLY for the source code.
//
// Copyright 2019 Thomas Weber
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// function definitions
int calcValueOfEventCombination(int calcTypeOfValueOfEventCombination, int numRandomVariables, int* valueCombination, int* valuesOfRandomVariables);
int _stdcall calcProbabilitiesOfGroupEvents_DLL(int* numGroupEvents, double* probOfGroupEvent, int* numRandomVariables, int* numValuesPerRandomVar, double* probOfRandomVariables, int* valuesOfRandomVariables, int* calcTypeOfValueOfEventCombination);

// Called by Excel
int _stdcall calcProbabilitiesOfGroupEvents_DLL(int* numGroupEvents, double* probOfGroupEvent, int* numRandomVariables, int* numValuesPerRandomVar, double* probOfRandomVariables, int* valuesOfRandomVariables, int* calcTypeOfValueOfEventCombination)
{
    // parameters ok?
    if (*numRandomVariables == 0 || *numGroupEvents == 0 || *numValuesPerRandomVar == 0) return 1;

	// constants
	double			minProb					= 0.00001;

    // locals
    double			curProb;
    int				curRandomVariable;
    int				curValueOfEventCombination;
    int*			eventCombination		= new int[*numRandomVariables];
    int*			valueCombination		= new int[*numRandomVariables];
	bool			allCombinationsTried	= false;
	
    // init
	for (curRandomVariable = 0; curRandomVariable < *numRandomVariables; curRandomVariable++) {
        eventCombination[curRandomVariable] = 1;
    }
	for (curValueOfEventCombination = 0; curValueOfEventCombination < *numGroupEvents; curValueOfEventCombination++) {
        probOfGroupEvent[curValueOfEventCombination] = 0;
    }

    // check all possible event combinations
	do {
        // calc current value of the event combination
        for (curRandomVariable = 0; curRandomVariable < *numRandomVariables; curRandomVariable++) {
             valueCombination[curRandomVariable] = valuesOfRandomVariables[curRandomVariable * (*numValuesPerRandomVar) + eventCombination[curRandomVariable] - 1];
        }
        curValueOfEventCombination = calcValueOfEventCombination(*calcTypeOfValueOfEventCombination, *numRandomVariables, valueCombination, valuesOfRandomVariables);
        
        // save probability of current event combination
        if (curValueOfEventCombination > 0 && curValueOfEventCombination <= *numGroupEvents) {
            curProb = 1;
             for (curRandomVariable = 0; curRandomVariable < *numRandomVariables; curRandomVariable++) {
                 curProb *= probOfRandomVariables[curRandomVariable * (*numValuesPerRandomVar) + eventCombination[curRandomVariable] - 1];
            }
            probOfGroupEvent[curValueOfEventCombination-1] += curProb;
        }
        
        // generate next event combination as long as any probability is lower than 'minProb'
        curRandomVariable = *numRandomVariables - 1;

		// increase last random variable to next minimum
		do {
			eventCombination[curRandomVariable]++;
			if (eventCombination[curRandomVariable] > *numValuesPerRandomVar) {
				break;
			}
		} while (probOfRandomVariables[curRandomVariable * (*numValuesPerRandomVar) + eventCombination[curRandomVariable] - 1] < minProb);

		while (eventCombination[curRandomVariable] > *numValuesPerRandomVar) {
			
			// leave if all combinations have been tested
			if (curRandomVariable == 0) {
				allCombinationsTried = true;
				break;
			}

			// set current considered random variable to next minimum
			eventCombination[curRandomVariable] = 1;
			while (probOfRandomVariables[curRandomVariable * (*numValuesPerRandomVar) + eventCombination[curRandomVariable] - 1] < minProb) {
				eventCombination[curRandomVariable]++;
				if (eventCombination[curRandomVariable] > *numValuesPerRandomVar) {
					return 2;	// error
				}
			}

			// consider next random variable
            curRandomVariable = curRandomVariable - 1;

			// increase random variable to next minimum
            do {
				eventCombination[curRandomVariable]++;
				if (eventCombination[curRandomVariable] > *numValuesPerRandomVar) {
					break;
				}
			} while (probOfRandomVariables[curRandomVariable * (*numValuesPerRandomVar) + eventCombination[curRandomVariable] - 1] < minProb);
        };

	} while (!allCombinationsTried);

	// free mem & return
	delete [] eventCombination;
	delete [] valueCombination;
	return 0;   
}

// This function simply sums up all values of an event combination
int calcValueOfEventCombination(int calcTypeOfValueOfEventCombination, int numRandomVariables, int* valueCombination, int* valuesOfRandomVariables)
{
    // locals
	int sumValue;
    int maxValue;
    int curValue;
    int curRandomVariable;
    int curRandomVariable2;
    int restingRandomVariables;
    int thresholdBasedOnRestingDices[5];

	// constants
	const int calcType_sum							= 0;
	const int calcType_sumEquals					= 1;
	const int calcType_doNotSumEquals				= 2;
	const int calcType_doNotSumEquals_throwSubseq	= 3;
	const int calcType_sumTwoLast					= 4;
	
    // get maximum
    maxValue = valueCombination[0];
	for (curRandomVariable = 1; curRandomVariable < numRandomVariables; curRandomVariable++) {
		if (valueCombination[curRandomVariable] > maxValue) {
            maxValue = valueCombination[curRandomVariable];
        }
    }

    // calc according to type
	switch (calcTypeOfValueOfEventCombination) 
	{
        // consider equal dices
        case calcType_sumEquals:
            for (curRandomVariable = 0; curRandomVariable < numRandomVariables - 1; curRandomVariable++) {
                curValue = valueCombination[curRandomVariable];
				for (curRandomVariable2 = curRandomVariable + 1; curRandomVariable2 < numRandomVariables; curRandomVariable2++) {
                    if (valueCombination[curRandomVariable] == valueCombination[curRandomVariable2]) {
                        curValue = curValue + valueCombination[curRandomVariable];
                    }
                }
                if (curValue > maxValue) {
                    maxValue = curValue;
                }
            }
            return maxValue;
            
        case calcType_doNotSumEquals:
            return maxValue;
        
        case calcType_doNotSumEquals_throwSubseq:
            // for subsequential throwing of dices.
            // throwing is stopped when the current value is greater than the threshold.
            thresholdBasedOnRestingDices[0] = valuesOfRandomVariables[2];
            thresholdBasedOnRestingDices[1] = valuesOfRandomVariables[2];
            thresholdBasedOnRestingDices[2] = valuesOfRandomVariables[3];
            thresholdBasedOnRestingDices[3] = valuesOfRandomVariables[4];
            thresholdBasedOnRestingDices[4] = valuesOfRandomVariables[4];
            for (curRandomVariable = 0; curRandomVariable < numRandomVariables; curRandomVariable++) {
                restingRandomVariables = numRandomVariables - curRandomVariable - 1;
                if (restingRandomVariables == 0 || valueCombination[curRandomVariable] > thresholdBasedOnRestingDices[restingRandomVariables - 1]) {
                    break;
                }
            }
            return valueCombination[curRandomVariable];
        
		case  calcType_sumTwoLast:
            // for subsequential throwing of dices, when the two last dices count
            // throwing is stopped when the current two last values is greater than the threshold.
			if (numRandomVariables > 1) {
				thresholdBasedOnRestingDices[0] = valuesOfRandomVariables[2];
				thresholdBasedOnRestingDices[1] = valuesOfRandomVariables[2];
				thresholdBasedOnRestingDices[2] = valuesOfRandomVariables[3];
				thresholdBasedOnRestingDices[3] = valuesOfRandomVariables[4];
				thresholdBasedOnRestingDices[4] = valuesOfRandomVariables[4];
				for (curRandomVariable = 1; curRandomVariable < numRandomVariables; curRandomVariable++) {
					restingRandomVariables = numRandomVariables - curRandomVariable - 1;
					if (restingRandomVariables == 0 || valueCombination[curRandomVariable-1] > thresholdBasedOnRestingDices[restingRandomVariables - 1]) {
						break;
					}
				}
                return valueCombination[curRandomVariable] + valueCombination[curRandomVariable-1];
			} else {
                return valueCombination[0];
            }

        case calcType_sum:
            sumValue = 0;
			for (curRandomVariable = 0; curRandomVariable < numRandomVariables; curRandomVariable++) {
                sumValue = sumValue + valueCombination[curRandomVariable];
            }
			return sumValue;
        
        default:
            //  MsgBox "ERROR: Invalid calculationType in function calcValueOfEventCombination()."
			return 0;
    }
    return 0;
}